// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v3.11.4
// source: github.com/moby/buildkit/util/stack/stack.proto

package stack

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type Stack struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Frames        []*Frame               `protobuf:"bytes,1,rep,name=frames,proto3" json:"frames,omitempty"`
	Cmdline       []string               `protobuf:"bytes,2,rep,name=cmdline,proto3" json:"cmdline,omitempty"`
	Pid           int32                  `protobuf:"varint,3,opt,name=pid,proto3" json:"pid,omitempty"`
	Version       string                 `protobuf:"bytes,4,opt,name=version,proto3" json:"version,omitempty"`
	Revision      string                 `protobuf:"bytes,5,opt,name=revision,proto3" json:"revision,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Stack) Reset() {
	*x = Stack{}
	mi := &file_github_com_moby_buildkit_util_stack_stack_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Stack) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Stack) ProtoMessage() {}

func (x *Stack) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_util_stack_stack_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Stack.ProtoReflect.Descriptor instead.
func (*Stack) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_util_stack_stack_proto_rawDescGZIP(), []int{0}
}

func (x *Stack) GetFrames() []*Frame {
	if x != nil {
		return x.Frames
	}
	return nil
}

func (x *Stack) GetCmdline() []string {
	if x != nil {
		return x.Cmdline
	}
	return nil
}

func (x *Stack) GetPid() int32 {
	if x != nil {
		return x.Pid
	}
	return 0
}

func (x *Stack) GetVersion() string {
	if x != nil {
		return x.Version
	}
	return ""
}

func (x *Stack) GetRevision() string {
	if x != nil {
		return x.Revision
	}
	return ""
}

type Frame struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Name          string                 `protobuf:"bytes,1,opt,name=Name,proto3" json:"Name,omitempty"`
	File          string                 `protobuf:"bytes,2,opt,name=File,proto3" json:"File,omitempty"`
	Line          int32                  `protobuf:"varint,3,opt,name=Line,proto3" json:"Line,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Frame) Reset() {
	*x = Frame{}
	mi := &file_github_com_moby_buildkit_util_stack_stack_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Frame) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Frame) ProtoMessage() {}

func (x *Frame) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_util_stack_stack_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Frame.ProtoReflect.Descriptor instead.
func (*Frame) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_util_stack_stack_proto_rawDescGZIP(), []int{1}
}

func (x *Frame) GetName() string {
	if x != nil {
		return x.Name
	}
	return ""
}

func (x *Frame) GetFile() string {
	if x != nil {
		return x.File
	}
	return ""
}

func (x *Frame) GetLine() int32 {
	if x != nil {
		return x.Line
	}
	return 0
}

var File_github_com_moby_buildkit_util_stack_stack_proto protoreflect.FileDescriptor

const file_github_com_moby_buildkit_util_stack_stack_proto_rawDesc = "" +
	"\n" +
	"/github.com/moby/buildkit/util/stack/stack.proto\x12\x05stack\"\x8f\x01\n" +
	"\x05Stack\x12$\n" +
	"\x06frames\x18\x01 \x03(\v2\f.stack.FrameR\x06frames\x12\x18\n" +
	"\acmdline\x18\x02 \x03(\tR\acmdline\x12\x10\n" +
	"\x03pid\x18\x03 \x01(\x05R\x03pid\x12\x18\n" +
	"\aversion\x18\x04 \x01(\tR\aversion\x12\x1a\n" +
	"\brevision\x18\x05 \x01(\tR\brevision\"C\n" +
	"\x05Frame\x12\x12\n" +
	"\x04Name\x18\x01 \x01(\tR\x04Name\x12\x12\n" +
	"\x04File\x18\x02 \x01(\tR\x04File\x12\x12\n" +
	"\x04Line\x18\x03 \x01(\x05R\x04LineB%Z#github.com/moby/buildkit/util/stackb\x06proto3"

var (
	file_github_com_moby_buildkit_util_stack_stack_proto_rawDescOnce sync.Once
	file_github_com_moby_buildkit_util_stack_stack_proto_rawDescData []byte
)

func file_github_com_moby_buildkit_util_stack_stack_proto_rawDescGZIP() []byte {
	file_github_com_moby_buildkit_util_stack_stack_proto_rawDescOnce.Do(func() {
		file_github_com_moby_buildkit_util_stack_stack_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_util_stack_stack_proto_rawDesc), len(file_github_com_moby_buildkit_util_stack_stack_proto_rawDesc)))
	})
	return file_github_com_moby_buildkit_util_stack_stack_proto_rawDescData
}

var file_github_com_moby_buildkit_util_stack_stack_proto_msgTypes = make([]protoimpl.MessageInfo, 2)
var file_github_com_moby_buildkit_util_stack_stack_proto_goTypes = []any{
	(*Stack)(nil), // 0: stack.Stack
	(*Frame)(nil), // 1: stack.Frame
}
var file_github_com_moby_buildkit_util_stack_stack_proto_depIdxs = []int32{
	1, // 0: stack.Stack.frames:type_name -> stack.Frame
	1, // [1:1] is the sub-list for method output_type
	1, // [1:1] is the sub-list for method input_type
	1, // [1:1] is the sub-list for extension type_name
	1, // [1:1] is the sub-list for extension extendee
	0, // [0:1] is the sub-list for field type_name
}

func init() { file_github_com_moby_buildkit_util_stack_stack_proto_init() }
func file_github_com_moby_buildkit_util_stack_stack_proto_init() {
	if File_github_com_moby_buildkit_util_stack_stack_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_util_stack_stack_proto_rawDesc), len(file_github_com_moby_buildkit_util_stack_stack_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   2,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_github_com_moby_buildkit_util_stack_stack_proto_goTypes,
		DependencyIndexes: file_github_com_moby_buildkit_util_stack_stack_proto_depIdxs,
		MessageInfos:      file_github_com_moby_buildkit_util_stack_stack_proto_msgTypes,
	}.Build()
	File_github_com_moby_buildkit_util_stack_stack_proto = out.File
	file_github_com_moby_buildkit_util_stack_stack_proto_goTypes = nil
	file_github_com_moby_buildkit_util_stack_stack_proto_depIdxs = nil
}
