// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v3.19.6
// source: noop.proto

package grpcpb

import (
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"

	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type NoopRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Ignored       string                 `protobuf:"bytes,1,opt,name=ignored,proto3" json:"ignored,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *NoopRequest) Reset() {
	*x = NoopRequest{}
	mi := &file_noop_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *NoopRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*NoopRequest) ProtoMessage() {}

func (x *NoopRequest) ProtoReflect() protoreflect.Message {
	mi := &file_noop_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use NoopRequest.ProtoReflect.Descriptor instead.
func (*NoopRequest) Descriptor() ([]byte, []int) {
	return file_noop_proto_rawDescGZIP(), []int{0}
}

func (x *NoopRequest) GetIgnored() string {
	if x != nil {
		return x.Ignored
	}
	return ""
}

type NoopResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Ignored       string                 `protobuf:"bytes,1,opt,name=ignored,proto3" json:"ignored,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *NoopResponse) Reset() {
	*x = NoopResponse{}
	mi := &file_noop_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *NoopResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*NoopResponse) ProtoMessage() {}

func (x *NoopResponse) ProtoReflect() protoreflect.Message {
	mi := &file_noop_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use NoopResponse.ProtoReflect.Descriptor instead.
func (*NoopResponse) Descriptor() ([]byte, []int) {
	return file_noop_proto_rawDescGZIP(), []int{1}
}

func (x *NoopResponse) GetIgnored() string {
	if x != nil {
		return x.Ignored
	}
	return ""
}

var File_noop_proto protoreflect.FileDescriptor

const file_noop_proto_rawDesc = "" +
	"\n" +
	"\n" +
	"noop.proto\x12\fio.podman.v1\"'\n" +
	"\vNoopRequest\x12\x18\n" +
	"\aignored\x18\x01 \x01(\tR\aignored\"(\n" +
	"\fNoopResponse\x12\x18\n" +
	"\aignored\x18\x01 \x01(\tR\aignored2E\n" +
	"\x04Noop\x12=\n" +
	"\x04Noop\x12\x19.io.podman.v1.NoopRequest\x1a\x1a.io.podman.v1.NoopResponseB'Z%go.podman.io/podman/v6/pkg/api/grpcpbb\x06proto3"

var (
	file_noop_proto_rawDescOnce sync.Once
	file_noop_proto_rawDescData []byte
)

func file_noop_proto_rawDescGZIP() []byte {
	file_noop_proto_rawDescOnce.Do(func() {
		file_noop_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_noop_proto_rawDesc), len(file_noop_proto_rawDesc)))
	})
	return file_noop_proto_rawDescData
}

var (
	file_noop_proto_msgTypes = make([]protoimpl.MessageInfo, 2)
	file_noop_proto_goTypes  = []any{
		(*NoopRequest)(nil),  // 0: io.podman.v1.NoopRequest
		(*NoopResponse)(nil), // 1: io.podman.v1.NoopResponse
	}
)

var file_noop_proto_depIdxs = []int32{
	0, // 0: io.podman.v1.Noop.Noop:input_type -> io.podman.v1.NoopRequest
	1, // 1: io.podman.v1.Noop.Noop:output_type -> io.podman.v1.NoopResponse
	1, // [1:2] is the sub-list for method output_type
	0, // [0:1] is the sub-list for method input_type
	0, // [0:0] is the sub-list for extension type_name
	0, // [0:0] is the sub-list for extension extendee
	0, // [0:0] is the sub-list for field type_name
}

func init() { file_noop_proto_init() }
func file_noop_proto_init() {
	if File_noop_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_noop_proto_rawDesc), len(file_noop_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   2,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_noop_proto_goTypes,
		DependencyIndexes: file_noop_proto_depIdxs,
		MessageInfos:      file_noop_proto_msgTypes,
	}.Build()
	File_noop_proto = out.File
	file_noop_proto_goTypes = nil
	file_noop_proto_depIdxs = nil
}
