// Code generated by go generate; DO NOT EDIT.
package system

import (
	"net/url"

	"go.podman.io/podman/v6/pkg/bindings/internal/util"
)

// Changed returns true if named field has been set
func (o *CheckOptions) Changed(fieldName string) bool {
	return util.Changed(o, fieldName)
}

// ToParams formats struct fields to be passed to API service
func (o *CheckOptions) ToParams() (url.Values, error) {
	return util.ToParams(o)
}

// WithQuick set field Quick to given value
func (o *CheckOptions) WithQuick(value bool) *CheckOptions {
	o.Quick = &value
	return o
}

// GetQuick returns value of field Quick
func (o *CheckOptions) GetQuick() bool {
	if o.Quick == nil {
		var z bool
		return z
	}
	return *o.Quick
}

// WithRepair set field Repair to given value
func (o *CheckOptions) WithRepair(value bool) *CheckOptions {
	o.Repair = &value
	return o
}

// GetRepair returns value of field Repair
func (o *CheckOptions) GetRepair() bool {
	if o.Repair == nil {
		var z bool
		return z
	}
	return *o.Repair
}

// WithRepairLossy set field RepairLossy to given value
func (o *CheckOptions) WithRepairLossy(value bool) *CheckOptions {
	o.RepairLossy = &value
	return o
}

// GetRepairLossy returns value of field RepairLossy
func (o *CheckOptions) GetRepairLossy() bool {
	if o.RepairLossy == nil {
		var z bool
		return z
	}
	return *o.RepairLossy
}

// WithUnreferencedLayerMaximumAge set field UnreferencedLayerMaximumAge to given value
func (o *CheckOptions) WithUnreferencedLayerMaximumAge(value string) *CheckOptions {
	o.UnreferencedLayerMaximumAge = &value
	return o
}

// GetUnreferencedLayerMaximumAge returns value of field UnreferencedLayerMaximumAge
func (o *CheckOptions) GetUnreferencedLayerMaximumAge() string {
	if o.UnreferencedLayerMaximumAge == nil {
		var z string
		return z
	}
	return *o.UnreferencedLayerMaximumAge
}
