// Code generated by go-swagger; DO NOT EDIT.

package network

// This file was generated by the swagger tool.
// Editing this file might prove futile when you re-run the swagger generate command

import (
	"net/netip"
)

// EndpointResource contains network resources allocated and used for a container in a network.
//
// swagger:model EndpointResource
type EndpointResource struct {

	// name
	// Example: container_1
	Name string `json:"Name"`

	// endpoint ID
	// Example: 628cadb8bcb92de107b2a1e516cbffe463e321f548feb37697cce00ad694f21a
	EndpointID string `json:"EndpointID"`

	// mac address
	// Example: 02:42:ac:13:00:02
	MacAddress HardwareAddr `json:"MacAddress"`

	// IPv4 address
	// Example: 172.19.0.2/16
	IPv4Address netip.Prefix `json:"IPv4Address"`

	// IPv6 address
	IPv6Address netip.Prefix `json:"IPv6Address"`
}
