// Code generated by "stringer -type=SizeB1024 -trimprefix=_i"; DO NOT EDIT.

package decor

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[_ib-1]
	_ = x[_iKiB-1024]
	_ = x[_iMiB-1048576]
	_ = x[_iGiB-1073741824]
	_ = x[_iTiB-1099511627776]
}

const (
	_SizeB1024_name_0 = "b"
	_SizeB1024_name_1 = "KiB"
	_SizeB1024_name_2 = "MiB"
	_SizeB1024_name_3 = "GiB"
	_SizeB1024_name_4 = "TiB"
)

func (i SizeB1024) String() string {
	switch {
	case i == 1:
		return _SizeB1024_name_0
	case i == 1024:
		return _SizeB1024_name_1
	case i == 1048576:
		return _SizeB1024_name_2
	case i == 1073741824:
		return _SizeB1024_name_3
	case i == 1099511627776:
		return _SizeB1024_name_4
	default:
		return "SizeB1024(" + strconv.FormatInt(int64(i), 10) + ")"
	}
}
